/* ---------------------------------------------------------------------------
 *      Mouse pointing device driver (line discipline)
 * ---------------------------------------------------------------------------
 *
 *	Copyright (C) 1997 by Jason McMullan
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 * 
 *      $Log: serialmouse.c,v $
 *      Revision 1.13  1997/10/19 22:18:33  becka
 *      Mouse numbering changed.
 *
 *      Revision 1.12  1997/09/12 07:28:47  taylorj
 *      Minor: Reformatted source to be consistent with GGI coding style guidelines,
 *      modified all debugging output to use the DEBUGx macros, and added CVS
 *      revision logging to headers where needed.
 *
 *      Revision 1.11  1997/05/20 03:59:11  ezrec
 *      Major: Went to a variable-packet-length scheme, should handle mouse-man
 *     	and microsoft mice properly now, along with mmseries mice.
 *
 *      Revision 1.10  1997/05/17 03:54:04  ezrec
 *      Minor: Fixed serialmouse button-release bug, played a few hours
 *     	worth of XQuake under Xggi 0.0.8, using the ggilib shim.
 *
 *      Revision 1.9  1997/05/14 06:04:41  ezrec
 *      Minor: Fixed serialmouse driver, and the xterm driver now handles
 *      ptr position
 *
 *      Revision 1.8  1997/05/08 04:24:54  ezrec
 *      Minor: Changes make top conform to the new system, 2.1.36 should work now
 *
 *      Revision 1.7  1997/05/06 15:50:45  ezrec
 *      Minor: Ray Glendenning - Removed `event' namespace pollution
 *
 *      Revision 1.6  1997/03/04 17:29:44  ezrec
 *      Minor: now works with the new 2.1.28 tty_ldisc structure
 *
 *      Revision 1.5  1997/02/21 14:20:44  ezrec
 *      Minor: M.G. found yet another bug in the 2.0.x code...
 *
 *      Revision 1.4  1997/02/21 02:07:10  ezrec
 *      Minor: Fixed bugaboos w/respect to 2.0.x code... (thanks M.G.!)
 *
 *      Revision 1.3  1997/02/20 00:35:18  ezrec
 *      Minor: Fixed font-mouse bug in vga.c, bad-string bug in ps2aux.c,
 *      2.0.x/2.1.x verify_area and include fixes, added 2.1.25,2.1.26
 *      patchability.
 *
 *      Revision 1.2  1997/02/19 11:39:10  seeger
 *      New install scheme, ps2aux.c moved, fixed bug in makefiles.
 *
 *      Revision 1.1  1997/02/08 20:52:44  ezrec
 *      Major: Added display selection - to get the 'default' display, use -1
 *
 *      Revision 1.3  1997/02/05 04:27:40  ezrec
 *      Major: Added support for Logitech and Microsoft mice
 *
 *      Revision 1.3  1997/02/05 04:18:21  ezrec
 *      Major: Added support for Microsoft and Logitech mouse types...
 *
 *      Revision 1.2  1997/02/04 04:31:07  ezrec
 *      Minor: Small bugfixes
 *
 *      Revision 1.1  1997/02/03 15:22:06  ezrec
 *      Minor: renamed msystems.c to serialmouse.c, as I will 
 *      adapt this code to be able to run multiple types of serial mice
 *      on multiple ports... And maybe even have a 'programmable' mouse
 *      packet parser... Who knows...
 *
 *      Revision 1.1  1997/02/03 14:54:04  ezrec
 *      Major: TTY line disciplline for serial mice. Currently only the
 *      MouseSystems mice are supported, but the rest will come later...
 *
 * ---------------------------------------------------------------------------
 */

#include <ggi/maintainers.h>
#define	MAINTAINER	Jason_McMullan
#define DRIVER_NAME	"Serial mouse driver"
#define DRIVER_REV	"$Revision: 1.13 $"

#include <linux/config.h>
#include <linux/module.h>

#include <ggi/kgi.h>
#include <ggi/serialmouse.h>

#include <asm/io.h>
#include <linux/random.h>

#if (LINUX_VERSION_CODE < LinuxVersionCode(2,1,0))
#  define NEED_VERIFY
#endif

#ifdef NEED_VERIFY
#  include <linux/mm.h>
#  include <asm/segment.h>
#else
#  include <asm/uaccess.h>	
#endif

typedef struct mouse_type_s 
{
	const char *manufacturer;
	const char *model;
	int packet_hdr_mask;
	int packet_hdr_id;
} mouse_type;

static mouse_type MouseType[MOUSE_TYPE_LAST] = 
{
	{
		"Microsoft",
		"Serial mouse",
		0x40,
		0x40
	},
	{
		"MouseSystems",
		"Serial mouse",
		0xF8,
		0x80
	},
	{
		"Logitech",
		"Serial mouse",
		0xe0,
		0x80
	},
};

typedef struct mouse_data_s 
{
	int magic;
	int id;
	enum mouse_id type;
	int count;
	int buttonstate;
	char buff[MOUSE_MAX_PACKET_LEN];
	struct kgi_input input_dev;
} mouse_data;

static mouse_data MouseDev[MOUSE_MAX_DEVS];

const unsigned int buttonmap[8]={
	0,
	0x2,
	0x4,
	0x6,
	0x1,
	0x3,
	0x5,
	0x7
};

/******************* mouse_data constructs *********************/
static void mouse_data_init(void)
{
	int i;
	for (i=0; i < MOUSE_MAX_DEVS; i++) {
		MouseDev[i].magic = ~MOUSE_MAGIC;
		MouseDev[i].id = i;
	}
}

static void mouse_data_set(mouse_data *dev, enum mouse_id mtype)
{
	dev->magic		= MOUSE_MAGIC;
	dev->id			= 0;
	dev->count		= 0;
	dev->type		= mtype;
	dev->buttonstate	= 0;
	dev->input_dev.next	= NULL;
	dev->input_dev.focus	= -1;
	dev->input_dev.id	= -1;
	strncpy(dev->input_dev.manufact, MouseType[mtype].manufacturer, 64);
	strncpy(dev->input_dev.model, MouseType[mtype].model, 64);
	dev->input_dev.events	= emPointer;
	dev->input_dev.report	= emPtrRelative | emPtrButton;
	dev->input_dev.ctrl	= NULL;
}

static mouse_data *mouse_data_alloc(enum mouse_id mtype, int display)
{
	int i;

	if (mtype < 0 || mtype >= MOUSE_TYPE_LAST)
	{
		
		return NULL;
	}

	for (i = 0; i < MOUSE_MAX_DEVS; i++) 
	{
		
		if (MouseDev[i].magic == ~MOUSE_MAGIC) {
			break;
		}

		if (i == MOUSE_MAX_DEVS) 
		{
			
			return NULL;
		}
		
		if (display < 0) 
		{
			
			MouseDev[i].id = i;
		}
		else
		{
			
			MouseDev[i].id = display;
		}
	}
	
	mouse_data_set(&(MouseDev[i]), mtype);
	return &(MouseDev[i]);
}


static void mouse_data_free(mouse_data *mdata)
{
	if (!mdata || mdata->magic != MOUSE_MAGIC)
	{
		
		return;
	}

	mdata->magic = ~MOUSE_MAGIC;
}

/******************* tty_struct operators *********************/

static int serialmouse_open_type_id(struct tty_struct *tty,
				    enum mouse_id mtype, int display)
{
	mouse_data *mdata = (mouse_data *)(tty->disc_data);
	int err;

	DEBUG2("Serialmouse_open_type_id: tty=%p, mtype=%d, dpy=%d",
	       tty, mtype, display);

	if (mdata && mdata->magic == MOUSE_MAGIC)
	{
		
		return -EEXIST;
	}

	if ((mdata = mouse_data_alloc(mtype, display)) == NULL) 
	{
		
		return -ENFILE;
	}

	if (tty->driver.flush_buffer)
	{
		
		tty->driver.flush_buffer(tty);
	}
	
	if (tty->ldisc.flush_buffer)
	{
		
		tty->ldisc.flush_buffer(tty);
	}
	
	tty->disc_data = mdata;
	
	DEBUG2("kgi_register_input: id=%d, dev=%p", mdata->id, 
	       &(mdata->input_dev));

	err = kgi_register_input(mdata->id, &(mdata->input_dev));
	if (err)
	{
		
		return err;
	}

	MOD_INC_USE_COUNT;

	return mdata->id;
}

static int serialmouse_open(struct tty_struct *tty)
{
	return serialmouse_open_type_id(tty, MOUSE_TYPE_DEFAULT, -1);
}

static void serialmouse_close(struct tty_struct *tty)
{
	mouse_data *mdata = (mouse_data *)(tty->disc_data);
	
	if (!mdata || mdata->magic != MOUSE_MAGIC)
	{
		
		return;
	}
	

	if (tty->driver.flush_buffer)
	{
		
		tty->driver.flush_buffer(tty);
	}
	
	if (tty->ldisc.flush_buffer)
	{
		
		tty->ldisc.flush_buffer(tty);
	}
	
	tty->disc_data = NULL;

	kgi_unregister_input(mdata->id, &(mdata->input_dev));

	mouse_data_free(mdata);

	MOD_DEC_USE_COUNT;
}

static int serialmouse_ioctl(struct tty_struct *tty, struct file *file,
			     unsigned int cmd, unsigned long arg)
{
	mouse_data *mdata = (mouse_data *)(tty->disc_data);
	enum mouse_id new_type,old_type;
	int new_id, old_id, err;
	struct serialmouse_attr *mattr = (struct serialmouse_attr *)(arg);

	if (!mdata || mdata->magic != MOUSE_MAGIC)
	{
		
		return -EEXIST;
	}

	if (_IOC_TYPE(cmd) != MOUSE_IOCTL_TYPE)
	{
		
		return -EINVAL;
	}
	
	
	switch (_IOC_NR(cmd)) 
	{
		
	case MOUSE_IOCTL_SETATTR:
		old_type = mdata->type;
		old_id = mdata->id;
#ifdef NEED_VERIFY
		if (verify_area(VERIFY_READ | VERIFY_WRITE, mattr,
				sizeof(struct serialmouse_attr)) != 0) 
		{
			
			return -EFAULT;
		}
		
		new_type = get_user(&(mattr->type));
		new_id = get_user(&(mattr->display));
#else
		if (get_user(new_type, &(mattr->type)) ||
		    get_user(new_id, &(mattr->display)))
		{
			
			return -EFAULT;
		}
#endif
		serialmouse_close(tty);
		err = serialmouse_open_type_id(tty, new_type, new_id);
		if (err)
		{
			
				return err;
		}
		
#ifdef NEED_VERIFY
		put_user(old_type, &(mattr->type));
		put_user(old_id, &(mattr->display));
#else
		if (put_user(old_type, &(mattr->type)) ||
		    put_user(old_id, &(mattr->display))) 
		{
			
			return -EFAULT;
		}
#endif
		return 0;
	case MOUSE_IOCTL_GETATTR:
#ifdef NEED_VERIFY
		if (verify_area(VERIFY_WRITE, mattr,
				sizeof(struct serialmouse_attr)) != 0) 
			return -EFAULT;
		put_user(mdata->type, &(mattr->type));
		put_user(mdata->id, &(mattr->display));
#else
		if (put_user(mdata->type, &(mattr->type)) ||
		    put_user(mdata->id, &(mattr->display))) 
		{
			
			return -EFAULT;
		}
#endif
		return 0;
	}
	
	return -EINVAL;
}
			

/* Parse packet, and modify mdata->count and mdtat->buf if
 * parse was incomplete
 */
static int packet_parse(mouse_data *mdata)
{
	int valid, used, buttons = 0, dx = 0, dy = 0;
	char *packet;
	unsigned char hdr_mask, hdr_id;
	ggi_event ev;

	if (!mdata || mdata->magic != MOUSE_MAGIC) 
	{
		DEBUG2("mdata=%p", mdata);
		return 0;
	}

	if (!KGI_IS_FOCUSED(mdata->input_dev)) {
		
		return 0;
	}

	if (mdata->count <= 0)
	{
		
		return 0;
	}
	

	hdr_mask = MouseType[mdata->type].packet_hdr_mask;
	hdr_id =MouseType[mdata->type].packet_hdr_id;
	packet = mdata->buff;

#if defined(DEBUG) && DEBUG
	printk("packet_parse(%d,{%.2x", mdata->count, packet[0]);
	for (used = 1; used < mdata->count; used++)
		printk(" %.2x", packet[used]);
	printk("})\n");
#endif

	valid = used = 0;
	switch (mdata->type) 
	{
	case MOUSE_TYPE_MOUSESYS:
		if ((packet[0] & hdr_mask) != hdr_id) 
		{
			DEBUG1("Invalid packet");
			used = 1;
			break;
		}
		if (mdata->count < 5)
		{
			
			break;
		}

		DEBUG2("Got mousesys packet");
		buttons = buttonmap[ (~packet[0]) & 0x07 ];
		dx= packet[1] + packet[3];
		dy= -(packet[2] + packet[4]);
		used = 5;
		valid = 1;
		break;
		
	case MOUSE_TYPE_MICROSOFT:
		if ((packet[0] & hdr_mask) == hdr_id) {
			if (mdata->count < 3)
			{
				
				break;
			}

			DEBUG2("Got ms packet");
			buttons = buttonmap[ ((packet[0] & 0x20) >> 3) |
				((packet[0] & 0x10) >> 4) ];
			dx = (char)(((packet[0] & 0x03) << 6) | 
				    (packet[1] & 0x3F));
			dy = (char)(((packet[0] & 0x0C) << 4) | 
				    (packet[2] & 0x3F));
			used = 3;
			valid = 1;
		} 
		else if ((packet[0] & ~0x23) == 0) 
		{
			DEBUG2("Got mouseman packet");
			buttons |= buttonmap[ (packet[0] & 0x20) >> 4 ];
			dx = dy = 0;
			used = 1;
			valid = 1;
		} else {
			DEBUG2("Invalid packet");
			used = 1;
		}
		break;
		
	case MOUSE_TYPE_LOGITECH:
		if ((packet[0] & hdr_mask) != hdr_id) {
			DEBUG2("Invalid packet");
			used = 1;
			break;
		}

		if (mdata->count < 3)
		{
			
			break;
		}
		
		DEBUG2("Got logitech packet");
		buttons = buttonmap[ packet[0] & 0x07 ];
		dx = (packet[0] & 0x10) ? packet[1] : -packet[1];
		dy = (packet[0] & 0x08) ? -packet[2] : packet[2];
		used = 3;
		valid = 1;
		break;
		
	default:
		INTERNALERROR;
		used = 1;
		break;
	}

	DEBUG2("Packet is %s, used %d bytes\n\n\n\n\n",
	       valid ? "valid" : "INVALID", used);

	if (used == 0)
	{
		
		return 0;
	}
	
	mdata->count -= used;
	if (mdata->count > 0) 
	{
		
		memcpy(packet, packet + used, mdata->count);
	}
	else
	{
		
		mdata->count = 0;
	}

	if (!valid) 
	{
		
		return 0;
	}

	ev.any.device = mdata->input_dev.id;
	ev.any.focus = mdata->input_dev.focus;
	ev.any.time = jiffies;

	if (dx || dy) 
	{
		ev.pmove.size = sizeof(ggi_pmove_event);

		ev.pmove.wheel = 0;
		ev.pmove.dummy[0] = ev.pmove.dummy[1] = 0;

		ev.pmove.x = dx;
		ev.pmove.y = dy;

		NOTICE2("evPtrRelative");

		ev.pmove.type = evPtrRelative;
		kgi_handle_input(&ev);
	}

	/* Change in button state?	
	 */

	ev.pbutton.size = sizeof(ggi_pbutton_event);
	ev.pbutton.type = evPtrButtonPress;
	if ( (ev.pbutton.button = (buttons ^ mdata->buttonstate) & buttons) ) {
		ev.pbutton.state = (mdata->buttonstate |= ev.pbutton.button);
		kgi_handle_input(&ev);
	}
	
	ev.pbutton.type = evPtrButtonRelease;
	ev.pbutton.state =  buttons;
	if ( (ev.pbutton.button = (buttons ^ mdata->buttonstate) & ~buttons) )
		kgi_handle_input(&ev);

	mdata->buttonstate = buttons;

	DEBUG2("packet_parse: packet parsed");

	return 1;
}

static void serialmouse_receive_buf(struct tty_struct *tty,
				    const unsigned char *charp,
				    char *flagp, int len)
{
	mouse_data *mdata = (mouse_data *)(tty->disc_data);
	int i;
	
	if (!mdata || mdata->magic != MOUSE_MAGIC) 
	{
		
		return;
	}

	for (i = 0; i < len; i++) {
		/* Watch out! We may have had a line break
		 * or a parity error, so check the line stats
		 */
		if (flagp[i] != TTY_NORMAL) 
		{
			mdata->count = 0;
			continue;
		}

		mdata->buff[mdata->count++] = charp[i];
	}

	while (packet_parse(mdata));
}

static int serialmouse_receive_room(struct tty_struct *tty)
{
	mouse_data *mdata = (mouse_data *)(tty->disc_data);
	
	if (!mdata || mdata->magic != MOUSE_MAGIC)
	{
		
		return 0;
	}

	return (MOUSE_MAX_PACKET_LEN - mdata->count);
}
	
int init_module(void)
{
	static struct tty_ldisc serialmouse_ldisc;
	
	memset(&serialmouse_ldisc, 0, sizeof(struct tty_ldisc));
	serialmouse_ldisc.magic		= TTY_LDISC_MAGIC;
#if (LINUX_VERSION_CODE >= LinuxVersionCode(2,1,28))
	serialmouse_ldisc.name		= "serialmouse";
#endif
	serialmouse_ldisc.num		= N_MOUSE;
	serialmouse_ldisc.flags		= 0;
	serialmouse_ldisc.open		= serialmouse_open;
	serialmouse_ldisc.close		= serialmouse_close;
	serialmouse_ldisc.ioctl		= serialmouse_ioctl;
	serialmouse_ldisc.receive_buf	= serialmouse_receive_buf;
	serialmouse_ldisc.receive_room	= serialmouse_receive_room;

	mouse_data_init();

	tty_register_ldisc(N_MOUSE, &serialmouse_ldisc);
	DEBUG1("GGI Mouse (N_MOUSE) tty line discipline enabled.");

	return 0;
}

void cleanup_module(void)
{	
	tty_register_ldisc(N_MOUSE, NULL);
	DEBUG1("GGI Mouse (N_MOUSE) tty line discipline removed");
}
