/* ---------------------------------------------------------------------------
**	Generic Monitor driver for fixed frequency monitors
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997 by Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
** 
**	$Log: monosync.c,v $
**	Revision 1.10  1997/12/31 16:39:38  seeger
**	minor polishing
**	
**	Revision 1.9  1997/12/30 22:09:05  grimrath
**	*** empty log message ***
**	
**	Revision 1.8  1997/12/14 21:56:12  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.7  1997/12/03 15:43:19  seeger
**	* increased error tolerance to support TI TVP 3026 PLL.
**	
**
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define MAINTAINER	Steffen_Seeger
#define	DRIVER_NAME	"Monosync monitor driver"
#define	DRIVER_REV	"$Revision: 1.10 $"
#define	DEBUG_LEVEL	1	/* do silent checking	*/

#ifdef DEBUG
#define NEED_MONITOR_SYMS
#endif

#include <ggi/module.h>

#define HPOS            0x1000
#define HNEG            0
#define VPOS            0x2000
#define VNEG            0

#define	MAX_HTIMINGS	16	/* MUST BE A POWER OF 2 !!	*/
#define	HTIMING_MASK	(MAX_HTIMINGS-1)

#include "monitors.inc"

#if HTIMINGS > MAX_HTIMINGS
#error more htimings than allowed by MAX_HTIMINGS
#endif

/*
**	KGI module interface
*/

int kgim_monitor_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);
	DEBUG1("%s %s (%dx%d pxl, %d x %d mm, %d-%d kHz, %d-%d Hz)",
	       monitor.manufact, monitor.model,
	       monitor.maxdots.x, monitor.maxdots.y,
	       monitor.size.x, monitor.size.y,
	       monitor.hfreq.min / 1000, monitor.hfreq.max / 1000, 
	       monitor.vfreq.min, monitor.vfreq.max);

	card->monitor = &monitor;
	return EOK;
}

void kgim_monitor_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

static int err(int pixels, int ns, struct kgi_mode *mode)
{
	int h = (mode->request.graphtype == GT_TEXT16)
		? 2*mode->request.text.x
		: 18; 	/* Allow timings to differ by up to 18 dclks */
	int width = (ns * (mode->dclk / 10000)) / 100000;

	int ret = abs(pixels - width) > h;

	if (ret) {

		DEBUG0("Timing wrong because %i - %i > %i", pixels, width, h);
	}

	return ret;
}

int kgim_monitor_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd *cmd)
{
	long v = 0, h = 0;
	ulong dclk;

	#define VISX   (mode->x.width)
	#define VISY   (mode->y.width)

	if ((VISX > (int)(monitor.maxdots.x)) || 
	    (VISY > (int)(monitor.maxdots.y))) {

		DEBUG2("Resolution too high (%ix%i).", VISX, VISY);
		return -E(MONITOR, UNKNOWN);
	}

	while ((v < VTIMINGS) && ((int)vtiming[v].width != VISY)) {

		v++;
	}

	if ((v == VTIMINGS) || (vtiming[v].width != VISY)) {

		DEBUG2("No vtiming for %i lines found.", VISY);
		return -E(MONITOR, UNKNOWN);
	}

	h = vtiming[v].polarity & HTIMING_MASK;
	dclk = (((ggi_uint) mode->x.width * 100000) / htiming[h].width) * 10000;

	switch (*cmd) {

	case CMD_PROPOSE:
		mode->dclk		= dclk;
		mode->y			= vtiming[v];
		mode->y.polarity	= (vtiming[v].polarity & VPOS) ? 1 : 0;

		#define POS(X)  (((int) VISX * htiming[h].X) / htiming[h].width)
		mode->x.width		= POS(width);
		mode->x.blankstart	= POS(blankstart);
		mode->x.syncstart	= POS(syncstart);
		mode->x.syncend		= POS(syncend);
		mode->x.blankend	= POS(blankend);
		mode->x.total		= POS(total);
		mode->x.polarity	= (vtiming[v].polarity & HPOS) ? 1 : 0;
		#undef POS
		  
		/* !!! Hack for the Mystique */
		if (!strcmp(dpy->card->model,"Mystique") && 
		    mode->request.graphtype == GT_8BIT) {
			mode->x.total     = (mode->x.total     + 15) & ~15;
			mode->x.syncstart = (mode->x.syncstart + 15) & ~15;
			mode->x.syncend   = (mode->x.syncend   + 15) & ~15;
		}
		
		if (monitor.flags & FL_MON_PROPSIZE) {

			mode->request.size.x = VISX * monitor.size.x
				/ monitor.maxdots.x;
			mode->request.size.y = VISY * monitor.size.y
				/ monitor.maxdots.y;
			
		} else {
				
			mode->request.size = monitor.size;
		}

		DEBUG2("propose dclk = %i Hz, hfreq = %i Hz, vfreq = %i Hz",
		       mode->dclk, mode->dclk / mode->x.total,
		       mode->dclk / (mode->x.total * mode->y.total));

		*cmd = CMD_LOWER;
		return EOK;

	case CMD_LOWER:

		if (mode->dclk < monitor.dclk.min) {

			return -E(MONITOR, UNKNOWN);
		}

		if (abs(mode->dclk - dclk) > (2 * dclk / 100)) {

			DEBUG2("dclk is %i (should %li) Hz", mode->dclk, dclk);
			return -E(MONITOR, UNKNOWN);
		}

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_RAISE:

		if (mode->dclk > monitor.dclk.max) {

			return -E(MONITOR, UNKNOWN);
		}

		if (abs(mode->dclk - dclk) > (2 * dclk / 100)) {

			DEBUG2("dclk is %i (should %li) Hz", mode->dclk, dclk);
			return -E(MONITOR, UNKNOWN);
		}

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_CHECK:

		if ((mode->dclk < monitor.dclk.min) || 
		    (monitor.dclk.max < mode->dclk)) {

			DEBUG2("DCLK of %i Hz out of bounds.", mode->dclk);
			return -E(MONITOR, UNKNOWN);
		}

		if ((monitor.maxdots.x < (int) VISX) || 
		    (monitor.maxdots.y < (int) VISY)) {

			DEBUG2("Resolution too high (%ix%i).", VISX, VISY);
			return -E(MONITOR, UNKNOWN);
		}

		if ((abs(mode->dclk - dclk) > (2 * dclk / 100)) || 
		    ((mode->dclk / mode->x.total) < monitor.hfreq.min) || 
		    ((mode->dclk / mode->x.total) > monitor.hfreq.max)) {

			DEBUG2("hfreq (%i Hz) out of bounds.",
			       mode->dclk / mode->x.total);
			return -E(MONITOR, UNKNOWN);
		}
	
		#define	ERR(X) (err(mode->x.X, htiming[h].X, mode) \
				|| (mode->y.X != vtiming[v].X))

		if (   ERR(width)   || ERR(blankstart) || ERR(syncstart) 
		    || ERR(syncend) || ERR(blankend)   || ERR(total)) {

			NOTICE2("Error in timings.");
			return -E(MONITOR, UNKNOWN);
		}
		#undef ERR

		*cmd = CMD_READY;
		return EOK;

	default:

		INTERNALERROR;
		return -E(MONITOR, UNKNOWN);
	}

	#undef VISX
	#undef VISY
}

void kgim_monitor_set_mode(struct kgi_display *dpy, struct kgi_mode *TM)
{
}

int kgim_monitor_command (struct kgi_graphic *state, ggi_uint cmd,ggi_uint iobuf)
{
#if 0
	switch (cmd) {

	case MONITOR_GETINFO:

		_out_buf = &monitor; 
		_io_cnt = sizeof(monitor);
		return EOK;

	case MONITOR_SETINFO:

		if (suser()) {

			memcpy(&monitor, _io_buf, sizeof(monitor));
			return EOK;

		} else {

			return -EPERM;
		}

	default:

		return -E(MONITOR, UNKNOWN);
	}
#endif
	return -E(MONITOR, UNKNOWN);
}
