/* --------------------------------------------------------------------------
**	A generic debugging framework.
** --------------------------------------------------------------------------
**
**	Copyright (C) 1997 by Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**
**	We allow for different debugging levels, with increasing strength of
**	checks and 'verbosity' of the program output. The following macros
**	are defined:
**
**	DEBUG_LEVEL (specified externally)
**	the current debug level. Use this to mark code with
**	"#ifdef DEBUG_LEVEL >= <level>" / "#endif" pairs that should only be
**	executed for certain levels and that is more than one statement.
**	You can also use TRACE<level>({ ... statements ... })
**
**	INTERNAL_ERROR;
**		Ooops. Mark code parts that should never be executed with this
**		macro. This will print an message where the error happened
**		and whom to report it.
**		
**	ERROR(format, args...);
**		print an error message with preceding "error: " string.
**		These messages will always appear somewhere, no matter
**		what DEBUG_LEVEL is.
**
**	NOTICE(format, args...);
**		Print an message that notifies the user about something.
**		This is intended to report useful info, e.g. detected stuff etc.
**		These messages may be useful in special cases (debugging),
**		but need not be cared about usually. These messages will
**		always appear somewhere, no matter what DEBUG_LEVEL is.
**
**	ASSERT(x);
**		Check if x is true. If so, do nothing, else print a message.
**		Checking will only be done if DEBUG_LEVEL is defined. For
**		released (non-beta) code this will do nothing. This is intended
**		for internal paranoia-checking, e.g. does the struct pointer
**		_really_ point to the data you assume it does point to?
**		example:	ASSERT(f->magic == 2);
**
**	PASSERT(x, fmt, args...);
**		Check if x is true. If so, do nothing, else print a verbose
**		message. Checking will only be done if DEBUG_LEVEL is defined.
**		This is intended for parameter checking of well-defined APIs.
**		example:	PASSERT(x == 2, "x is %i\n", x);
**
**	DEBUG<level>(fmt, args...);
**		Print a message that gives debugging info, but only if
**		DEBUG_LEVEL is greater than <level>. The message will be 
**		prefixed with the current source file name and line number.
**		example:	DEBUG3("test %i %x\n", i, x);
**
**	TRACE<level>(x);
**		Execute the statement x if the DEBUG_LEVEL defined not less
**		than <level>.
**		example:	TRACE(memset(mydata, 0x32, sizeof(*mydata));
**		NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE	NOTE
**		NOTE							NOTE
**		NOTE    TAKE CARE x DOESN'T MODIFY PROGRAM STATE	NOTE
**		NOTE (EXCEPT YOU **KNOW**EXACTLY** WHAT YOU ARE DOING)	NOTE
**		NOTE							NOTE
**		NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE	NOTE
**
**	One final advice: USE THESE MACROS! You may find them annoying, but
**	they will by sure save you hours of bug hunting and turn out to be
**	the more useful the more complex your code becomes!
*/

#ifndef	_DEBUG_H
#define	_DEBUG_H

#ifdef	__KERNEL__ /* { */

	#define	__DL0	KERN_ERR
	#define	__DL1	KERN_ERR
	#define	__DL2	KERN_ERR
	#define	__DL3	KERN_ERR

	#define	__DEBUG_MSG(level, format, args...) \
		printk( level "%s:%i: " format "\n" , \
			__FILE__ , __LINE__ , ##args )

	#define	INTERNAL_ERROR	\
		printk( KERN_ERR "%s:%i: internal error. Please report to %s." \
			" Thank you!\n" , __FILE__ , __LINE__ , MAINTAINER)

	#define	ERROR(format, args...) \
		printk( KERN_ERR "error: " format "\n" , ##args )

	#define	NOTICE(format, args...)	\
		printk( KERN_ERR format "\n" , ##args )
#else /* } { */

	#define	__DL0	""
	#define	__DL1	""
	#define	__DL2	""
	#define	__DL3	""

#if defined(sun) /* { */
	#define	__DEBUG_MSG(level, format, args) \
		fprintf(stderr, level "%s:%i: " format "\n" , \
			__FILE__ , __LINE__ , ##args )

	#define	INTERNAL_ERROR	\
		fprintf(stderr , "%s:%i: internal error. Please report to %s." \
			" Thanks.\n" , __FILE__ , __LINE__ , MAINTAINER )

	#define	ERROR(format, args)	\
		fprintf(stderr , "error: " format "\n" , ##args )

	#define	NOTICE(format, args)	\
		fprintf(stderr , fmt , ##args );
#elif defined (__sgi) /* } { */
	extern void debug_printf (char *form, ...);
	extern void ERROR (char *format, ...);
	extern void PASSERT (int x, char *arg, ...);
	extern void DEBUG0 (char *arg, ...);
	extern void DEBUG1 (char *arg, ...);
	extern void DEBUG2 (char *arg, ...);
	extern void DEBUG3 (char *arg, ...);

	#define NOTICE(x) debug_printf ( x )

	#define	__DEBUG_MSG(level, format, args) \
		debug_printf(level "%s:%i: " format "\n" , \
			__FILE__ , __LINE__ , ##args )

	#define	INTERNAL_ERROR	\
		debug_printf("%s:%i: internal error. Please report to %s." \
			" Thanks.\n" , __FILE__ , __LINE__ , MAINTAINER )
#else /* } { */
	#define	__DEBUG_MSG(level, format, args...) \
		fprintf(stderr , level "%s:%i: " format "\n" , \
			__FILE__ , __LINE__ , ##args )

	#define	INTERNAL_ERROR	\
		fprintf(stderr , "%s:%i: internal error. Please report to %s." \
			" Thanks.\n" , __FILE__ , __LINE__ , MAINTAINER )

	#define	ERROR(format, args...)	\
		fprintf(stderr , "error: " format "\n" , ##args )

	#define	NOTICE(format, args...)	\
		fprintf(stderr , fmt , ##args);
#endif /* } */
#endif /* } */


#ifndef DEBUG_LEVEL  /* { */

#if defined(sun) /* { */
	#define	ASSERT(x)		do { } while (0)
	void PASSERT(int x, char *arg, ...) { do { } while (0); }

	#define	TRACE0(x)		do { } while (0)
	void DEBUG0(char *arg, ...) { do { } while (0); }

	#define	TRACE1(x)		do { } while (0)
	void DEBUG1(char *arg, ...) { do { } while (0); }

	#define	TRACE2(x)		do { } while (0)
	void DEBUG2(char *arg, ...) { do { } while (0); }

	#define	TRACE3(x)		do { } while (0)
	void DEBUG3(char *arg, ...) { do { } while (0); }
#elif !defined (__sgi) /* } { */
	#define	ASSERT(x)		do { } while (0)
	#define	PASSERT(x, args...)	do { } while (0)

	#define	TRACE0(x)		do { } while (0)
	#define	DEBUG0(args...)		do { } while (0)

	#define	TRACE1(x)		do { } while (0)
	#define	DEBUG1(args...)		do { } while (0)

	#define	TRACE2(x)		do { } while (0)
	#define	DEBUG2(args...)		do { } while (0)

	#define	TRACE3(x)		do { } while (0)
	#define	DEBUG3(args...)		do { } while (0)
#endif /* } */

#else /* } { */

	#if DEBUG_LEVEL < 0  /* { */

		#error	Invalid DEBUG_LEVEL
	#endif  /* } */

	/*	Level 0 is meant to do 'silent' checking of internal
	**	state. No debugging messages except for obvious wrong use of
	**	well defined interfaces. Use PASSERT to check and report
	**	the wrong parameters. All assertions are checked and 
	**	TRACE0() statements are executed. This is meant to do checking
	**	of internal state (not that we are paranoid, but ... :) and easy
	**	detection of possibly wrong or non-performant use of the public
	**	interfaces.
	*/
	#define	ASSERT(x) \
		if (x) {} else { \
			__DEBUG_MSG(__DL0, "assertion %s failed.", #x); }

#if defined(sun) /* { */
	void PASSERT(int x, char *arg, ...) {
		if (!x) {
			__DEBUG_MSG(__DL0, "assertion %s failed.", x);
			fprintf (stderr, arg..);
			__DEBUG_MSG(__DL0, ##arg, args);
			/* XXX implement the rest */
		}
	}
#elif !defined (__sgi) /* } { */
	#define	PASSERT(x, args...) \
		if (x) {} else { \
			__DEBUG_MSG(__DL0, "assertion %s failed.", #x); \
			__DEBUG_MSG(__DL0, ##args); }
#endif /* } */

	#define	TRACE0(x)	x

#if defined(sun) /* { */
	void DEBUG0(char *arg, ...) {
		__DEBUG_MSG(__DL0, arg);
		/* XXX finish this */
	}
#elif !defined(__sgi) /* } { */
	#define	DEBUG0(args...)	__DEBUG_MSG(__DL0, ##args)
#endif /* } */

	/*	Level 1 should be a bit more verbose, e.g. what is going on
	**	inside your code.
	*/
	#if DEBUG_LEVEL > 0 /* { */
		#define	TRACE1(x)	x
	 #if defined(sun) /* { */
		void DEBUG1(char *arg, ...) {
			__DEBUG_MSG(__DL1, arg);
			/* XXX finish this ... */
		}
	 #elif !defined (__sgi) /* } { */
		#define DEBUG1(args...)	__DEBUG_MSG(__DL1, ##args)
	 #endif /* }*/
	#else /* } { */
		#define	TRACE1(x)	do { } while (0)
	 #if defined(sun) /* { */
		void DEBUG1(char *arg, ...) { do { } while (0); }
	 #elif !defined (__sgi) /* } { */
		#define	DEBUG1(args...)	do { } while (0)
	 #endif /* } */
	#endif /* } */

	/*	Level 2 and higher should be even more verbose about internal
	**	state, what is right, what is wrong, etc.
	*/
	#if DEBUG_LEVEL > 1 /* { */
		#define	TRACE2(x)	x
	 #if defined(sun) /* { */
		void DEBUG2(char *arg, ...) {
			__DEBUG_MSG(__DL2, char *arg);
			/* XXX finish this ... */
		}
	 #elif !defined (__sgi) /* } { */
		#define	DEBUG2(args...)	__DEBUG_MSG(__DL2, ##args)
	 #endif /* } */
	#else /* } { */
		#define	TRACE2(x)	do { } while (0)
	 #if defined(sun) /* { */
		void DEBUG2(char *arg, ...) { do { } while (0); }
	 #elif !defined (__sgi) /* } { */
		#define	DEBUG2(args...)	do { } while (0)
         #endif /* } */
	#endif /* } */

	/*	Level 3 and higher should put out lowest level actions. E.g.
	**	track I/O operations, etc. May be really slow, but it's ok
	**	as long as you find the bugs.
	*/
	#if DEBUG_LEVEL > 2 /* { */
		#define	TRACE3(x)	x
	 #if defined(sun) /* { */
		void DEBUG3(char *arg, ...) {
			__DEBUG_MSG(__DL3, arg);
			/* XXX finish this ... */
		}
	 #elif !defined (__sgi) /* } { */
		#define	DEBUG3(args...)	__DEBUG_MSG(__DL3, ##args)
	 #endif /* } */
	#else /* } { */
		#define	TRACE3(x)	do { } while (0)
	 #if defined(sun) /* { */
		void DEBUG3(char *arg, ...) {
			__DEBUG_MSG(__DL3, arg);
			/* XXX finish this ... */
		}
	 #elif !defined (__sgi) /* } { */
		#define DEBUG3(args...)	do { } while (0)
	 #endif /* } */
	#endif /* } */
#endif  /* } */

/*	These are for backward compatibility. Clean your sources and don't use
**	them anymore. They will dissapear soon.
*/

#define	NOTICE0(x)	DEBUG0( x )
#define	NOTICE1(x)	DEBUG1( x )
#define	NOTICE2(x)	DEBUG2( x )
#define	NOTICE3(x)	DEBUG3( x )

#endif	/* #ifdef _DEBUG_H */
